import boto3
import copy
import hashlib
import logging
import json
import time
import typing
import uuid

from bert import \
    encoders as bert_encoders, \
    datasource as bert_datasource, \
    constants as bert_constants

from datetime import datetime, timedelta

logger = logging.getLogger(__name__)
PWN = typing.TypeVar('PWN')
DELAY: int = 15

class QueueItem:
    __slots__ = ('_payload', '_identity')
    _payload: typing.Dict[str, typing.Any]
    _identity: str
    def __init__(self: PWN, payload: typing.Dict[str, typing.Any], identity: str = None) -> None:
        self._payload = payload
        self._identity = identity

    def calc_identity(self: PWN) -> str:
        if self._identity:
            return self._identity

        combined: str = ''.join(bert_encoders.encode_identity_object(self._payload))
        combined: str = f'{combined}-{uuid.uuid4()}'
        return hashlib.sha256(combined.encode(bert_constants.ENCODING)).hexdigest()

    def keys(self: PWN) -> typing.Any:
        return super(QueueItem, self).keys()

    def get(self: PWN, name: str, default: typing.Any = None) -> typing.Any:
        return self._payload.get(name, default)

    def clone(self: PWN) -> typing.Any:
        return self.__class__(copy.deepcopy(self._payload))

    def __getitem__(self: PWN, name: str) -> typing.Any:
        try:
            return self._payload[name]
        except KeyError:
            raise KeyError(f'key-name[{name}] not found')

    def __setitem__(self: PWN, name: str, value: typing.Any) -> None:
        self._payload[name] = value

    def __delitem__(self: PWN, name: str) -> None:
        try:
            del self._payload[name]
        except KeyError:
            raise KeyError(f'key-name[{name}] not found')

class BaseQueue:
    _table_name: str
    _value: QueueItem
    def __init__(self: PWN, table_name: str) -> None:
        self._table_name = table_name
        self._value = None

    def __next__(self) -> typing.Any:
        if not self._value is None:
            logger.info('Destroying Value')
            self._destroy(self._value)
            self._value = None

        self._value = self.get()
        if self._value is None or self._value == 'STOP':
            raise StopIteration

        return self._value

    def get(self: PWN) -> QueueItem:
        raise NotImplementedError

    def put(self: PWN, value: typing.Union[typing.Dict[str, typing.Any], QueueItem]) -> None:
        raise NotImplementedError

    def __iter__(self) -> PWN:
        return self

    def _destroy(self: PWN, queue_item: QueueItem) -> None:
        raise NotImplementedError

class DynamodbQueue(BaseQueue):
    _dynamodb_client: 'boto3.client("dynamodb")'
    def __init__(self: PWN, table_name: str) -> None:
        super(DynamodbQueue, self).__init__(table_name)
        self._dynamodb_client = boto3.client('dynamodb')

    def _destroy(self: PWN, queue_item: QueueItem, confirm_delete: bool = False) -> None:
        if confirm_delete:
            self._dynamodb_client.delete_item(
                TableName=self._table_name,
                Key={'identity': {'S': queue_item.calc_identity()}},
                Expected={'identity': {'Exists': True, 'Value': value['identity']}})

        else:
            self._dynamodb_client.delete_item(
                TableName=self._table_name,
                Key={'identity': {'S': queue_item.calc_identity()}})

    def put(self: PWN, value: typing.Union[typing.Dict[str, typing.Any], QueueItem]) -> None:
        if isinstance(value, dict):
            queue_item = QueueItem(value)

        elif isinstance(value, QueueItem):
            queue_item = value

        else:
            raise NotImplementedError

        encoded_value = bert_encoders.encode_object({
            'identity': queue_item.calc_identity(),
            'datum': queue_item.clone(),
        })
        self._dynamodb_client.put_item(TableName=self._table_name, Item=encoded_value)

    def get(self: PWN) -> typing.Dict[str, typing.Any]:
        try:
            value: typing.Any = self._dynamodb_client.scan(TableName=self._table_name, Select='ALL_ATTRIBUTES', Limit=1)['Items'][0]
        except IndexError:
            return None

        else:
            queue_item = QueueItem(bert_encoders.decode_object(value['datum']), value['identity']['S'])
            if value['identity']['S'] in ['sns-entry', 'invoke-arg', 'api-gateway']:
                return queue_item

            # The order of data when coming out of the database maynot be preserved, resulting in a different identity
            # assert queue_item.calc_identity() == value['identity']['S'], f'{queue_item.calc_identity()} != {value["identity"]["S"]}'
            return queue_item

class RedisQueue(BaseQueue):
    _table_name: str
    _redis_client: 'redis-client'
    def __init__(self, table_name: str) -> None:
        super(RedisQueue, self).__init__(table_name)
        self._redis_client = bert_datasource.RedisConnection.ParseURL(bert_constants.REDIS_URL).client

    def flushdb(self) -> None:
        self._redis_client.flushdb()

    def _destroy(self: PWN, queue_item: QueueItem) -> None:
        pass

    def get(self) -> QueueItem:
        try:
            value: str = self._redis_client.lpop(self._table_name).decode(bert_constants.ENCODING)
        except AttributeError:
            return 'STOP'

        else:
            return bert_encoders.decode_object(json.loads(value)['datum'])

    def put(self: PWN, value: typing.Dict[str, typing.Any]) -> None:
        encoded_value = json.dumps(bert_encoders.encode_object({
            'identity': 'local-queue',
            'datum': value
        })).encode(bert_constants.ENCODING)
        self._redis_client.rpush(self._table_name, encoded_value)


class StreamingQueue(DynamodbQueue):
    """
    When deploying functions to AWS Lambda, auto-invocation is available as an option to run the functions. With StreamingQueue, we want to push local objects into
        the available API already utilized. We also want to keep the available `put` function so that the `done_queue` api will still push contents into the next `work_queue`.
        We'll also argment the local `get` function api and only pull from records local to the stream and not pull from dynamodb.
    """
    # Share the memory across invocations, within the same process/thread. This allows for
    #   comm_binders to be called multipule-times and still pull from the same queue
    _queue: typing.List[typing.Dict[str, typing.Any]] = []
    def local_put(self: PWN, record: typing.Union[typing.Dict[str, typing.Any], QueueItem]) -> None:
        if isinstance(record, dict):
            queue_item = QueueItem(bert_encoders.decode_object(record['datum']), record['identity']['S'])

        elif isinstance(record, QueueItem):
            queue_item = record

        self._queue.append(queue_item)

    def get(self: PWN) -> QueueItem:
        try:
            value: QueueItem = self._queue.pop(0)
        except IndexError:
            return super(StreamingQueue, self).get()

        else:
            return value

class LocalQueue(DynamodbQueue):
    """
    When testing, its convenient to use only a LocalQueue
    """
    _key: str = None
    # Share the memory across invocations, within the same process/thread. This allows for
    #   comm_binders to be called multipule-times and still pull from the same queue
    _queue: typing.List[typing.Dict[str, typing.Any]] = []
    def __init__(self: PWN, key: str, pipeline_type: bert_constants.PipelineType) -> None:
        self._key = key

    def local_put(self: PWN, record: typing.Dict[str, typing.Any]) -> None:
        self._queue.append(copy.deepcopy(record))

    def put(self: PWN, record: typing.Dict[str, typing.Any]) -> None:
        logger.info(f'LocalQueue Put[{record}]')

    def get(self: PWN) -> typing.Dict[str, typing.Any]:
        try:
            # may need to unpack because local queues are used for debugging in AWS Lambda
            value: typing.Any = self._queue.pop(0)
        except IndexError:
            return None

        else:
            return value




import logging
import typing

logger = logging.getLogger('') # <--- Probable a good idea to name your logger. '' is the 'root' logger
sysHandler = logging.StreamHandler()
sysHandler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
logger.addHandler(sysHandler)
logger.setLevel(logging.INFO)




import hashlib
import typing
import types

from bert import constants

def calc_func_space(func: typing.Union[str, types.FunctionType]) -> str:
    if isinstance(func, str):
        combined: str = hashlib.sha1(func.encode(constants.ENCODING)).hexdigest()
        return f'bert-etl-{combined}'

    elif isinstance(func, types.FunctionType):
        combined: str = hashlib.sha1(func.__name__.encode(constants.ENCODING)).hexdigest()
        return f'bert-etl-{combined}'

    else:
        raise NotImplementedError

def calc_func_key(*args: typing.List[str]) -> str:
    combined: str = hashlib.sha1(''.join(args).encode(constants.ENCODING)).hexdigest()
    return f'bert-etl-{combined}'




import json
import types
import typing

from bert import constants

from datetime import datetime

PWN = typing.TypeVar('PWN')

class IdentityEncoder(json.JSONEncoder):
    def default(self: PWN, obj: typing.Any) -> typing.Any:
        if isinstance(obj, datetime):
            return obj.strftime(constants.DATETIME_FORMAT)

        elif hasattr(datum, '_payload') and datum.__class__.__name__ == 'QueueItem':
            return super(IdentityEncoder, self).default(datum._payload)

        return super(IdentityEncoder, self).default(obj)

def _find_aws_encoding(datum: typing.Any) -> typing.Dict[str, typing.Any]:
    if isinstance(datum, dict):
        return 'M'

    elif isinstance(datum, list):
        return 'L'

    elif isinstance(datum, bytes):
        return 'B'

    elif isinstance(datum, str):
        return 'S'

    elif isinstance(datum, bool):
        return 'S'

    elif isinstance(datum, int):
        return 'S'

    elif isinstance(datum, float):
        return 'S'

    elif hasattr(datum, '_payload') and datum.__class__.__name__ == 'QueueItem':
        return 'M'

    elif datum is None:
        return 'S'

    raise NotImplementedError(f'Unable to encode[{type(datum)}] datum[{datum}]')

def encode_aws_object(datum: typing.Any) -> typing.Dict[str, typing.Any]:
    if isinstance(datum, dict):
        for key, value in datum.items():
            datum[key] = {_find_aws_encoding(value): encode_aws_object(value)}

        return datum

    elif isinstance(datum, (list, tuple, types.GeneratorType)):
        for idx, value in enumerate(datum):
            datum[idx] = {_find_aws_encoding(value): encode_aws_object(value)}

        return datum

    elif isinstance(datum, bool):
        return f'bool:{datum}'

    elif isinstance(datum, int):
        return f'int:{datum}'

    elif isinstance(datum, float):
        return f'float:{datum}'

    elif isinstance(datum, str):
        return datum

    elif hasattr(datum, '_payload') and datum.__class__.__name__ == 'QueueItem':
        return encode_aws_object(datum._payload)

    elif datum is None:
        return 'null:'

    else:
        raise NotImplementedError

    return None

def decode_aws_object(datum: typing.Dict[str, typing.Any]) -> typing.Any:
    for encoding_type, encoded in datum.items():
        if encoding_type == 'M':
            for key, value, in encoded.items():
                encoded[key] = decode_aws_object(value)

            return encoded

        elif encoding_type == 'L':
            for idx, value in enumerate(encoded):
                encoded[idx] = decode_aws_object(value)

            return encoded

        elif encoding_type == 'B':
            return encoded

        elif encoding_type == 'S' and encoded[:5] == 'bool:':
            value: str = encoded.split(':', 1)[1].lower()
            if value == 'true':
                return True

            elif value == 'false':
                return False

            raise NotImplementedError(f'Unable to decode datum[{value}]')

        elif encoding_type == 'S' and encoded[:4] == 'int:':
            return int(encoded.split(':', 1)[1])

        elif encoding_type == 'S' and encoded[:6] == 'float:':
            return float(encoded.split(':', 1)[1])

        elif encoding_type == 'S' and encoded[:5] == 'null:':
            return None

        elif encoding_type == 'S':
            return encoded

        return None




import collections
import importlib
import logging
import multiprocessing
import os
import subprocess
import time
import types
import typing

from bert import \
    queues as bert_queues, \
    constants as bert_constants, \
    shortcuts as bert_shortcuts, \
    encoders as bert_encoders, \
    exceptions as bert_exceptions

from bert import naming as bert_naming

ZIP_EXCLUDES: typing.List[str] = [
    '*.exe', '*.DS_Store', '*.Python', '*.git', '.git/*', '*.zip', '*.tar.gz',
    '*.hg', 'pip', 'docutils*', 'setuputils*', '__pycache__/*',
]
COMMON_EXCLUDES: typing.List[str] = ['env', 'lambdas']

logger = logging.getLogger(__name__)

def scan_jobs(module_name: str) -> typing.Dict[str, typing.Any]:
    jobs: typing.Dict[str, typing.Any] = {}
    module = importlib.import_module(f'{module_name}.jobs')
    for member_name in dir(module):
        if member_name.startswith('_'):
            continue

        member = getattr(module, member_name)
        if type(member) != types.FunctionType:
            continue

        if not hasattr(member, 'done_key') or not hasattr(member, 'work_key'):
            continue

        jobs[member_name] = member

    # Order the jobs correctly
    ordered = collections.OrderedDict()
    while len(ordered.keys()) != len(jobs.keys()):
        if len(ordered.keys()) == 0:
            for job_name, job in jobs.items():
                if job.parent_func == 'noop':
                    ordered[job_name] = job
                    break

            else:
                raise NotImplementedError(f'NoopSpace not found')

        else:
            latest: types.FunctionType = [item for item in ordered.values()][-1]
            for job_name, job in jobs.items():
                if job.parent_func == latest:
                    ordered[job_name] = job
                    break
    return ordered

def map_jobs(jobs: typing.Dict[str, typing.Any], module_name: str) -> None:
    confs: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
    bert_configuration = bert_shortcuts.load_configuration() or {}
    deployment_config = bert_shortcuts.obtain_deployment_config(bert_configuration)

    # Validate we have access to objects in AWS
    bert_shortcuts.head_bucket_for_existance(deployment_config.s3_bucket)

    for job_name, job in jobs.items():
        identity_encoders: typing.List[str] = bert_shortcuts.merge_lists(
            bert_configuration.get('every_lambda', {'identity_encoders': []}).get('identity_encoders', []),
            bert_configuration.get(job_name, {'identity_encoders': []}).get('identity_encoders', []),
            ['bert.encoders.base.IdentityEncoder'])

        queue_encoders: typing.List[str] = bert_shortcuts.merge_lists(
            bert_configuration.get('every_lambda', {'queue_encoders': []}).get('queue_encoders', []),
            bert_configuration.get(job_name,        {'queue_encoders': []}).get('queue_encoders', []),
            ['bert.encoders.base.encode_aws_object'])

        queue_decoders: typing.List[str] = bert_shortcuts.merge_lists(
            bert_configuration.get('every_lambda', {'queue_decoders': []}).get('queue_decoders', []),
            bert_configuration.get(job_name,        {'queue_decoders': []}).get('queue_decoders', []),
            ['bert.encoders.base.decode_aws_object'])

        invoke_args: typing.List[str] = bert_shortcuts.merge_lists(
            bert_configuration.get('every_lambda', {'invoke_args': []}).get('invoke_args', []),
            bert_configuration.get(job_name,        {'invoke_args': []}).get('invoke_args', []),
            [])
        invoke_args: typing.List[typing.Dict[str, typing.Any]] = bert_shortcuts.load_invoke_args(invoke_args)

        # This ENVVar will exist when the bert-etl-monitor function is deployed to AWS Lambda. bert-etl-monitor is a 
        #  utility function and doesn't require any additional packages, encoders or decoders.
        if not os.environ.get('BERT_MODULE_NAME', None) is None:
            identity_encoders = []
            queue_encoders = []
            queue_decoders = []

        # Make sure the encoders exist
        bert_encoders.load_encoders_or_decoders(identity_encoders)
        bert_encoders.load_encoders_or_decoders(queue_encoders)
        bert_encoders.load_encoders_or_decoders(queue_decoders)

        # concurrency_limit is checked against AWS account execution limit in bert.deploy.utils
        concurrency_limit: int = bert_shortcuts.get_if_exists('concurrency_limit', '0', int, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
        runtime: int = bert_shortcuts.get_if_exists('runtime', 'python3.6', str, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
        memory_size: int = bert_shortcuts.get_if_exists('memory_size', '128', int, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
        if int(memory_size / 64) != memory_size / 64:
            raise bert_exceptions.BertConfigError(f'MemorySize[{memory_size}] must be a multiple of 64')
        min_proced_items: int = bert_shortcuts.get_if_exists('min_proced_items', '25', int, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))

        # iam
        iam_execution_role_arn: str = bert_shortcuts.get_if_exists(
            'iam.execution_role_arn', None, str,
            bert_configuration.get('every_lambda', {'iam':{}}),
            bert_configuration.get(job_name, {'iam': {}}))

        # kms
        kms_alias: str = bert_shortcuts.get_if_exists(
            'kms.alias', None, str,
            bert_configuration.get('every_lambda', {'kms': {}}),
            bert_configuration.get(job_name, {'kms': {}}))

        kms_usernames: typing.List[str] = bert_shortcuts.get_and_merge_if_exists(
            'kms.usernames', None, list,
            bert_configuration.get('every_lambda', {'kms': {}}),
            bert_configuration.get(job_name, {'kms': {}}))

        invoke_args: typing.List[typing.Dict[str, typing.Any]] = bert_shortcuts.load_invoke_args(invoke_args)
        # events
        # sns topic to proc lambda
        sns_topic_arn: str = bert_shortcuts.get_if_exists(
            'events.sns_topic_arn', None, str,
            bert_configuration.get('every_lambda', {'events':{}}),
            bert_configuration.get(job_name, {'events': {}}))

        # schedule_expression will be validated before executing the deploy script
        schedule_expression: str = bert_shortcuts.get_if_exists(
            'events.schedule_expression', None, str,
            bert_configuration.get('every_lambda', {'events':{}}),
            bert_configuration.get(job_name, {'events': {}}))
        bottle_schedule_expression: str = bert_shortcuts.get_if_exists(
            'events.bottle_schedule_expression', 'rate(5 minutes)', str,
            bert_configuration.get('every_lambda', {'events':{}}),
            bert_configuration.get(job_name, {'events': {}}))

        api_stage: str = bert_shortcuts.get_if_exists(
            'api.stage', None, str,
            bert_configuration.get('every_lambda', {'api': {}}),
            bert_configuration.get(job_name, {'api': {}}))

        api = getattr(job, '_api', None)
        if api is None:
            api_path: str = f'{job_name}.json'
            api_method: str = 'get'

        else:
            api_path: str = api.route.route.strip('/')
            api_method: str = api.method.value

        batch_size: int = bert_shortcuts.get_if_exists('batch_size', '100', int,
            bert_configuration.get('every_lambda', {}),
            bert_configuration.get(job_name, {}))
        batch_size_delay: int = bert_shortcuts.get_if_exists('batch_size_delay', '0', int,
            bert_configuration.get('every_lambda', {}),
            bert_configuration.get(job_name, {}))

        timeout: int = bert_shortcuts.get_if_exists('timeout', '900', int, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
        env_vars: typing.Dict[str, str] = bert_shortcuts.merge_env_vars(
            bert_configuration.get('every_lambda', {'environment': {}}).get('environment', {}),
            bert_configuration.get(job_name, {'environment': {}}).get('environment', {}))

        # Set QueueType to dynamodb unless they've specifically requested a BERT_QUEUE_TYPE
        env_vars['BERT_QUEUE_TYPE'] = env_vars.get('BERT_QUEUE_TYPE', 'dynamodb')
        env_vars['BERT_MODULE_NAME'] = module_name
        requirements: typing.Dict[str, str] = bert_shortcuts.merge_requirements(
            bert_configuration.get('every_lambda', {'requirements': {}}).get('requirements', {}),
            bert_configuration.get(job_name, {'requirements': {}}).get('requirements', {}))

        if not api_stage is None:
            job_handler = f'{job.func_space}.{job_name}_api_handler'

        elif job.pipeline_type is bert_constants.PipelineType.BOTTLE and job.parent_noop_space == False:
            job_handler = f'{job.func_space}.{job_name}_manager'

        else:
            job_handler = f'{job.func_space}.{job_name}_handler'

        confs[job_name] = {
                'job': job,
                'deployment': {key: value for key, value in deployment_config._asdict().items()},
                'aws-deployed': {
                    'events': {},
                    'bottle': {},
                    'iam': {},
                    'kms': {},
                },
                'aws-deploy': {
                    'timeout': timeout,
                    'runtime': runtime,
                    'memory-size': memory_size, # must be a multiple of 64, increasing memory size also increases cpu allocation
                    'requirements': requirements,
                    'handler': job_handler,
                    'lambda-name': job_name,
                    'work-table-name': job.work_key,
                    'done-table-name': job.done_key,
                    'environment': env_vars,
                    'batch-size': batch_size,
                    'batch-size-delay': batch_size_delay,
                    'concurrency-limit': concurrency_limit,
                    'invoke-args': invoke_args,
                },
                'aws-build': {
                    'lambdas-path': os.path.join(os.getcwd(), 'lambdas'),
                    'excludes': ZIP_EXCLUDES + COMMON_EXCLUDES,
                    'path': os.path.join(os.getcwd(), 'lambdas', job_name),
                    'archive-path': os.path.join(os.getcwd(), 'lambdas', f'{job_name}.zip')
                },
                'runner': {
                    'environment': env_vars,
                    'max-retries': 10,
                },
                'iam': {
                    'execution-role-arn': iam_execution_role_arn,
                },
                'kms': {
                    'alias': kms_alias,
                    'usernames': kms_usernames,
                },
                'events': {
                    'schedule-expression': schedule_expression,
                    'sns-topic-arn': sns_topic_arn,
                    # "cron(0 20 * * ? *)" or "rate(5 minutes)"
                    # 'rate': 'rate(5 minutes)',
                },
                'api': {
                    'stage': api_stage,
                    'name': f'{job_name}-rest-api',
                    'path': api_path,
                    'method': api_method,
                },
                'bottle': {
                    'schedule-expression': bottle_schedule_expression,
                },
                'spaces': {
                    'func_space': job.func_space,
                    'work-key': job.work_key,
                    'done-key': job.done_key,
                    'pipeline-type': job.pipeline_type,
                    'workers': job.workers,
                    'scheme': job.schema,
                    'parent': {
                        'noop-space': job.parent_noop_space,
                        'space': job.parent_space,
                        'work-key': job.parent_func_work_key,
                        'done-key': job.parent_func_done_key,
                    },
                    'min_proced_items': min_proced_items
                },
                'encoding': {
                    'identity_encoders': identity_encoders,
                    'queue_encoders': queue_encoders,
                    'queue_decoders': queue_decoders,
                }
            }
        pass

    return confs

def run_command(cmd: str, allow_error: typing.List[int] = [0]) -> str:
    ologger = logging.getLogger('.'.join([__name__, multiprocessing.current_process().name]))
    cmd: typing.List[str] = cmd.split(' ')
    proc = subprocess.Popen(' '.join(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    while proc.poll() is None:
        time.sleep(.1)

    if proc.poll() > 0:
        if not proc.poll() in allow_error:
            raise NotImplementedError(f'{proc.poll()}, {proc.stderr.read()}')

        return proc.stderr.read().decode(bert_constants.ENCODING)

    return proc.stdout.read().decode(bert_constants.ENCODING)


def comm_binders(func: types.FunctionType) -> typing.Tuple['QueueType', 'QueueType', 'ologger']:
    ologger = logging.getLogger('.'.join([func.__name__, multiprocessing.current_process().name]))
    ologger.debug(f'Bert Queue Type[{bert_constants.QueueType}]')
    if bert_constants.QueueType is bert_constants.QueueTypes.Dynamodb:
        return bert_queues.DynamodbQueue(func.work_key), bert_queues.DynamodbQueue(func.done_key), ologger

    elif bert_constants.QueueType is bert_constants.QueueTypes.StreamingQueue:
        return bert_queues.StreamingQueue(func.work_key), bert_queues.StreamingQueue(func.done_key), ologger

    elif bert_constants.QueueType is bert_constants.QueueTypes.LocalQueue:
        return bert_queues.LocalQueue(func.work_key), bert_queues.LocalQueue(func.done_key), ologger

    elif bert_constants.QueueType is bert_constants.QueueTypes.Redis:
        return bert_queues.RedisQueue(func.work_key), bert_queues.RedisQueue(func.done_key), ologger

    else:
        raise NotImplementedError(f'Unsupported QueueType[{bert_constants.QueueType}]')


def flush_db():
    if bert_constants.QueueType in [
        bert_constants.QueueType.Dynamodb,
        bert_constants.QueueType.StreamingQueue]:
        raise NotImplementedError(f'Flush Dynamodb Tables')

    elif bert_constants.QueueType is bert_constants.QueueType.Redis:
        import redis
        from bert import constants, datasource
        redis_connection: datasource.RedisConnection = datasource.RedisConnection.ParseURL(constants.REDIS_URL)
        logger.info(f'Flushing Redis DB[{redis_connection.db}]')
        redis_connection.client.flushdb()

    else:
        raise NotImplementedError(constants.QueueType)





import base64
import json
import types
import typing

import numpy as np

from bert.encoders import base as base_encoders

PWN = typing.TypeVar('PWN')

class NumpyIdentityEncoder(base_encoders.IdentityEncoder):
    def default(self: PWN, obj: typing.Any) -> typing.Any:
        if isinstance(obj, (
            np.float32, np.float64, np.float16,
            np.complex64,
            np.int8, np.int16, np.int32, np.int64,
            np.uint8, np.uint16, np.uint32, np.uint64,
            np.intc, np.intp,
            np.bool_,
            np.ndarray)):
            return base64.b64encode(obj.tostring(order='F')).decode('ascii')

        return super(NumpyIdentityEncoder, self).default(obj)

def _find_aws_encoding(datum: typing.Any) -> typing.Dict[str, typing.Any]:
    if isinstance(datum, (
        np.float32, np.float64, np.float16,
        np.complex64,
        np.int8, np.int16, np.int32, np.int64,
        np.uint8, np.uint16, np.uint32, np.uint64,
        np.intc, np.intp,
        np.bool_,
        np.ndarray)):
        return 'S'

    return base_encoders._find_aws_encoding(datum)

def encode_aws_object(datum: typing.Any) -> typing.Dict[str, typing.Any]:
    if isinstance(datum, dict):
        for key, value in datum.items():
            datum[key] = {_find_aws_encoding(value): encode_aws_object(value)}

        return datum

    elif isinstance(datum, (list, tuple, types.GeneratorType)):
        for idx, value in enumerate(datum):
            datum[idx] = {_find_aws_encoding(value): encode_aws_object(value)}

        return datum

    # Floats
    elif isinstance(datum, np.float32):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.float32:{encoded_value}'

    elif isinstance(datum, np.float64):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.float64:{encoded_value}'

    elif isinstance(datum, np.float16):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.float16:{encoded_value}'

    # Complex
    elif isinstance(datum, np.complex64):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.complex64:{encoded_value}'

    # Ints
    elif isinstance(datum, np.int8):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.int8:{encoded_value}'

    elif isinstance(datum, np.int16):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.int16:{encoded_value}'

    elif isinstance(datum, np.int32):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.int32:{encoded_value}'

    elif isinstance(datum, np.int64):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.int64:{encoded_value}'

    # Uint
    elif isinstance(datum, np.uint8):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.uint8:{encoded_value}'

    elif isinstance(datum, np.uint16):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.uint16:{encoded_value}'

    elif isinstance(datum, np.uint32):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.uint32:{encoded_value}'

    elif isinstance(datum, np.uint64):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.uint64:{encoded_value}'

    # intc, intp, bool
    elif isinstance(datum, np.intc):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.intc:{encoded_value}'

    elif isinstance(datum, np.intp):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.incp:{encoded_value}'

    elif isinstance(datum, np.bool_):
        encoded_value = base64.b64encode(datum.tostring()).decode('ascii')
        return f'np.bool_:{encoded_value}'

    elif isinstance(datum, np.ndarray):
        shape = str(datum.shape).strip('()')
        shape = ','.join(shape.split(', '))
        dtype = datum.dtype.type.__name__
        encoded_value = base64.b64encode(datum.tostring(order='F')).decode('ascii')
        return f'np.ndarray:{shape}:{dtype}:{encoded_value}'

    return base_encoders.encode_aws_object(datum)

def decode_aws_object(datum: typing.Dict[str, typing.Any]) -> typing.Any:
    try:
        datum.items()
    except Exception as err:
        import ipdb; ipdb.set_trace()
        pass
    for encoding_type, encoded in datum.items():
        if encoding_type == 'M':
            for key, value, in encoded.items():
                encoded[key] = decode_aws_object(value)

            return encoded

        elif encoding_type == 'L':
            for idx, value in enumerate(encoded):
                encoded[idx] = decode_aws_object(value)

            return encoded

        elif encoding_type == 'S' and encoded[:11] == 'np.float16:':
            binary = base64.b64decode(encoded[11:])
            return np.fromstring(binary, dtype=np.float16)[0]

        elif encoding_type == 'S' and encoded[:11] == 'np.float32:':
            binary = base64.b64decode(encoded[11:])
            return np.fromstring(binary, dtype=np.float32)[0]

        elif encoding_type == 'S' and encoded[:11] == 'np.float64:':
            binary = base64.b64decode(encoded[11:])
            return np.fromstring(binary, dtype=np.float64)[0]

        # np.complex64
        elif encoding_type == 'S' and encoded[:12] == 'np.complex64:':
            binary = base64.b64decode(encoded[12:])
            return np.fromstring(binary, dtype=np.complex64)[0]

        # np.int*
        elif encoding_type == 'S' and encoded[:8] == 'np.int8:':
            binary = base64.b64decode(encoded[8:])
            return np.fromstring(binary, dtype=np.int8)[0]

        elif encoding_type == 'S' and encoded[:9] == 'np.int16:':
            binary = base64.b64decode(encoded[9:])
            return np.fromstring(binary, dtype=np.int16)[0]

        elif encoding_type == 'S' and encoded[:9] == 'np.int32':
            binary = base64.b64decode(encoded[9:])
            return np.fromstring(binary, dtype=np.int32)[0]

        elif encoding_type == 'S' and encoded[:9] == 'np.int64':
            binary = base64.b64decode(encoded[9:])
            return np.fromstring(binary, dtype=np.int64)[0]

        # np.uint*
        elif encoding_type == 'S' and encoded[:9] == 'np.uint8:':
            binary = base64.b64decode(encoded[9:])
            return np.fromstring(binary, dtype=np.uint8)[0]

        elif encoding_type == 'S' and encoded[:10] == 'np.uint16:':
            binary = base64.b64decode(encoded[10:])
            return np.fromstring(binary, dtype=np.uint8)[0]

        elif encoding_type == 'S' and encoded[:10] == 'np.uint32:':
            binary = base64.b64decode(encoded[10:])
            return np.fromstring(binary, dtype=np.uint32)[0]

        elif encoding_type == 'S' and encoded[:10] == 'np.uint64:':
            binary = base64.b64decode(encoded[10:])
            return np.fromstring(binary, dtype=np.uint64)[0]

        # np.intc/p
        elif encoding_type == 'S' and encoded[:8] == 'np.intc:':
            binary = base64.b64decode(encoded[8:])
            return np.fromstring(binary, dtype=np.intc)[0]

        elif encoding_type == 'S' and encoded[:8] == 'np.intp:':
            binary = base64.b64decode(encoded[8:])
            return np.fromstring(binary, dtype=np.intp)[0]

        elif encoding_type == 'S' and encoded[:11] == 'np.ndarray:':
            shape, dtype, value = datum[11:].split(':', 2)
            dtype_value = {
                'int8': np.int8,
                'int16': np.int16,
                'int32': np.int32,
                'int64': np.int64,
                'float16': np.float16,
                'float32': np.float32,
                'float64': np.float64,
                'complex': np.complex,
                'complex64': np.complex64,
                'uint8': np.uint8,
                'uint16': np.uint16,
                'unit32': np.uint32,
                'uint64': np.uint64,
                'intc': np.intc,
                'intp': np.intp,
                'bool': np.bool_,
                'bool_': np.bool_
            }.get(dtype, None)
            if dtype_value is None:
                raise NotImplementedError(f'np.ndarray dtype[{dtype}] is not supported. Open an issue request for support: https://github.com/jbcurtin/bert-etl/issues')

            shape: typing.List[int] = [vec for vec in map(lambda x: int(x), shape.split(','))]
            value: bytes = base64.b64decode(value)
            # finding some way to tell this function we're using Fortran order would be nice to have
            return np.fromstring(value, dtype=dtype_value).reshape(*shape)

    return base_encoders.decode_aws_object(datum)




import boto3
import collections
import importlib
import json
import logging
import os
import types
import typing
import yaml

from bert import \
    exceptions as bert_exceptions

from botocore.errorfactory import ClientError

from json.decoder import JSONDecodeError

logger = logging.getLogger(__name__)

def head_bucket_for_existance(bucket_name: str) -> None:
    if bucket_name is None:
        return None

    client = boto3.client('s3')
    try:
        client.head_bucket(Bucket=bucket_name)
    except ClientError as err:
        if '(403)' in err.args[0]:
            raise bert_exceptions.AWSError(f'Bucket[{bucket_name}] name is taken by someone else')

def obtain_secrets_config(bert_configuration: typing.Any) -> collections.namedtuple:
    defaults = {
        'key_alias': 'bert-etl',
        'usernames': [],
    }
    items = bert_configuration.get('secrets', {})
    for key, value in defaults.items():
        if not key in items.keys():
            items[key] = value

    keys = [key for key in items.keys()]
    values = []
    secrets_tuple = collections.namedtuple('Secrets', keys)
    for key in keys:
        values.append(items[key])

    return secrets_tuple(*values)

def obtain_deployment_config(bert_configuration: typing.Any) -> collections.namedtuple:
    """
    Written verbosely to allow for default_keys
    """
    defaults = {
        's3_bucket': None,
    }
    items = bert_configuration.get('deployment', {})
    for key, value in defaults.items():
        if not key in items.keys():
            items[key] = value

    keys = [key for key in items.keys()]
    values = []
    deployment_tuple = collections.namedtuple('Deployment', keys)
    for key in keys:
        values.append(items[key])

    return deployment_tuple(*values)

def save_configuration(configuration: typing.Dict[str, typing.Any]) -> None:
    conf_path: str = os.path.join(os.getcwd(), 'bert-etl.yaml')
    logger.info(f'Writing configuration to path[{conf_path}]')
    with open(conf_path, 'w', encoding='utf-8') as stream:
        stream.write(yaml.dump(configuration, indent=4, canonical=False))

def load_configuration() -> typing.Dict[str, typing.Any]:
    conf_path: str = os.path.join(os.getcwd(), 'bert-etl.yaml')
    if not os.path.exists(conf_path):
        logger.info(f"Configuration Path[{conf_path}] doesn't exist")
        return None

    with open(conf_path, 'r', encoding='utf-8') as stream:
        return yaml.load(stream.read(), Loader=yaml.FullLoader)

def get_and_merge_if_exists(keypath: str, default: typing.Any, object_type: typing.Any, defaults: typing.List[typing.Any], primary: typing.List[str]) -> typing.List[str]:
    steps: typing.List[str] = keypath.split('.')
    if len(steps) > 1:
        for key in steps[:-1]:
            try:
                if isinstance(defaults[key], (dict, list)):
                    defaults = defaults[key]

                else:
                    raise bert_exceptions.BertConfigError(f'Invalid KeyPath for Defaults: keypath[{keypath}], key[{key}]')

            except KeyError:
                pass

            try:
                if isinstance(primary[key], (dict, list)):
                    primary = primary[key]

                else:
                    raise bert_exceptions.BertConfigError(f'Invalid KeyPath for Primary: keypath[{keypath}], key[{key}]')

            except KeyError:
                pass
                # raise bert_exceptions.BertConfigError(f'Invalid KeyPath: keypath[{keypath}], key[{key}]')

    values: typing.List[str] = defaults.get(steps[-1], [])
    values.extend(primary.get(steps[-1], []))
    return values


def merge_lists(main: typing.List[typing.Any], secondary: typing.List[typing.Any], defaults: typing.List[typing.Any]) -> typing.List[typing.Any]:
    if len(main) == 0 and len(secondary) == 0:
        return defaults

    # Preserve order or list
    merged = []
    for value in secondary:
        if value in merged:
            continue

        merged.append(value)

    for value in main:
        if value in merged:
            continue

        merged.append(value)

    return merged

def merge_env_vars(defaults: typing.Dict[str, str], env_vars: typing.Dict[str, str]) -> typing.Dict[str, str]:
    merged: typing.Dict[str, str] = {key: value for key, value in defaults.items()}
    for key, value in env_vars.items():
        merged[key] = value

    return merged

def merge_requirements(defaults: typing.List[str], requirements: typing.List[str]) -> typing.List[str]:
    merged: typing.List[str] = {value for value in defaults}
    for value in requirements:
        if value in merged:
            continue

        merged.append(value)

    return [item for item in merged]

def get_if_exists(keypath: str, default: typing.Any, data_type: typing.Any, defaults: typing.Dict[str, str], input_vars: typing.Dict[str, str]) -> typing.Dict[str, str]:
    steps: typing.List[str] = keypath.split('.')
    if len(steps) > 1:
        for key in steps[:-1]:
            try:
                if isinstance(input_vars[key], (dict, list)):
                    input_vars = input_vars[key]

                else:
                    raise bert_exceptions.BertConfigError(f'Invalid KeyPath: keypath[{keypath}], key[{key}]')

            except KeyError:
                pass
                # raise bert_exceptions.BertConfigError(f'Invalid KeyPath: keypath[{keypath}], key[{key}]')

    key: str = steps[-1]
    try:
        return data_type(input_vars[key])
    except KeyError:
        pass

    except ValueError:
        raise bert_exceptions.BertConfigError(f'Key[{keypath}] is not DataType[{data_type}]')

    try:
        result: typing.Any = defaults.get(key, default)
        if result is None:
            return None

        return data_type(result)
    except ValueError:
        raise bert_exceptions.BertConfigError(f'Default Key[{keypath}] is not DataType[{data_type}]')


def _load_invoke_args_module(member_route: str) -> typing.Dict[str, typing.Any]:
    try:
        module_path, member_name = member_route.rsplit('.', 1)
        module = importlib.import_module(module_path)
    except ValueError:
        raise ImportError

    else:
        member: typing.Union[typing.Dict[str, typing.Any], types.FunctionType, None] = getattr(module, member_name, None)
        if member is None:
            raise ImportError

        elif isinstance(member, dict):
            return member

        elif isinstance(member, types.FunctionType):
            return member()

        else:
            raise NotImplementedError(member)

def load_invoke_args(invoke_args: typing.List[str]) -> typing.List[typing.Dict[str, typing.Any]]:
    loaded: typing.List[typing.Dict[str, typing.Any]] = []
    for invoke_item in invoke_args:
        if isinstance(invoke_item, dict):
            # yaml
            loaded.append(invoke_item)

        elif invoke_item.endswith('.json'):
            with open(invoke_item, 'r') as stream:
                loaded.append(json.loads(stream.read()))

        elif invoke_item.endswith('.yaml') or invoke_item.endswith('.yml'):
            with open(invoke_item, 'r') as stream:
                loaded.append(yaml.load(stream.read(), Loader=yaml.FullLoader))

        else:
            try:
                loaded.extend(_load_invoke_args_module(invoke_item))
            except ImportError:
                pass
            else:
                continue

            try:
                datum = json.loads(invoke_item)
                if isinstance(datum, list):
                    loaded.extend(datum)

                elif isinstance(datum, dict):
                    loaded.append(datum)

                else:
                    raise bert_exceptions.BertConfigError(f'Invalid Json Type. Only List, Dict are supported.')
            except JSONDecodeError:
                pass

            else:
                continue

            raise bert_exceptions.BertConfigError(f'Unable to load invoke_arg[{invoke_item}]')

    return loaded

def load_if_exists(keypath: str, configuration: typing.Dict[str, typing.Any]) -> typing.Any:
    steps = keypath.split('.')
    if len(steps) > 1:
        for key in steps[:-1]:
            try:
                if isinstance(configuration[key], (list, dict)):
                    configuration= configuration[key]

                else:
                    raise bert_exceptions.BertConfigError(f'Invalid Keypath: keypath[{keypath}]')
            except KeyError:
                pass

    try:
        return configuration[steps[-1]]
    except (KeyError, IndexError) as err:
        raise bert_exceptions.BertConfigError(f'Invalid Keypath: keypath[{keypath}]')

def write_keypath_value(keypath: str, configuration: typing.Dict[str, typing.Any], value: typing.Any) -> None:
    steps = keypath.split('.')
    if len(steps) > 1:
        for key in steps[:-1]:
            try:
                if isinstance(configuration[key], (list, dict)):
                    configuration = configuration[key]

                else:
                    raise bert_exceptions.BertConfigError(f'Invalid Keypath: keypath[{keypath}]')

            except KeyError:
                pass

    configuration[steps[-1]] = value

